# pyfoam\pyfoam\core\solvers\cfds\navier_stocks_solver.py

import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spla
from typing import Callable, Optional
from meshing.mesh import StructuredMesh
from conditions.condition import DirichletCondition, NeumannCondition

# =============================================================================
# 1. Navier–Stokes Solver via Finite Volume Method (FVM)
# =============================================================================

class NavierStokesFVM:
    def __init__(self,
                 mesh: StructuredMesh,
                 nu: float,
                 penalty: float,
                 f_u: Callable[[np.ndarray], float],
                 f_v: Callable[[np.ndarray], float],
                 dirichlet_bc_u: DirichletCondition,
                 dirichlet_bc_v: DirichletCondition,
                 neumann_bc_u: Optional[NeumannCondition] = None,
                 neumann_bc_v: Optional[NeumannCondition] = None):
        """
        Initialize the FVM solver for the steady Navier–Stokes equations (penalty formulation).

        Parameters:
          mesh              : StructuredMesh instance.
          nu                : Kinematic viscosity.
          penalty           : Penalty parameter (λ = 1/ε) to enforce near-incompressibility.
          f_u, f_v          : Source term functions for the x- and y-momentum equations.
          dirichlet_bc_u,v  : Dirichlet BC objects for u and v.
          neumann_bc_u,v    : Optional Neumann BC objects for u and v.
        """
        self.mesh = mesh
        self.nu = nu
        self.penalty = penalty
        self.f_u = f_u
        self.f_v = f_v
        self.dirichlet_bc_u = dirichlet_bc_u
        self.dirichlet_bc_v = dirichlet_bc_v
        self.neumann_bc_u = neumann_bc_u
        self.neumann_bc_v = neumann_bc_v

        # For FVM, the unknowns are cell averages.
        self.cell_centers, self.cell_volumes, self.grid_shape = self._compute_cell_info()
        self.num_cells = self.cell_centers.shape[0]
        # We'll store the velocity field as a concatenated vector [u; v].
        self.solution = None

    def _compute_cell_info(self):
        """Compute cell centers and volumes for a structured 2D mesh."""
        if not hasattr(self.mesh, 'num_cells'):
            raise ValueError("Mesh must have 'num_cells' attribute.")
        n_x, n_y = self.mesh.num_cells
        x_min, x_max = self.mesh.bounds[0]
        y_min, y_max = self.mesh.bounds[1]
        dx = (x_max - x_min) / n_x
        dy = (y_max - y_min) / n_y
        centers = []
        volumes = []
        for i in range(n_x):
            for j in range(n_y):
                center = np.array([x_min + (i + 0.5)*dx, y_min + (j + 0.5)*dy])
                centers.append(center)
                volumes.append(dx*dy)
        return np.array(centers), np.array(volumes), (n_x, n_y)

    def _assemble_operator(self, grid_shape, h):
        """
        Assemble a 5-point stencil matrix for the Laplacian on a uniform grid.
        This is used for each velocity component.
        
        The discrete operator is:
            L u_P = u_E + u_W + u_N + u_S - 4 u_P
        multiplied by (ν + penalty*h^2) factor later.
        Returns a sparse matrix L of shape (N, N).
        """
        n_x, n_y = grid_shape
        N = n_x * n_y
        data = []
        rows = []
        cols = []
        def lin_index(i, j):
            return i * n_y + j
        for i in range(n_x):
            for j in range(n_y):
                idx = lin_index(i, j)
                # For boundary cells, we set a placeholder (to be overridden by BC).
                if i == 0 or i == n_x - 1 or j == 0 or j == n_y - 1:
                    data.append(1.0)
                    rows.append(idx)
                    cols.append(idx)
                else:
                    # Interior cell: five-point stencil.
                    data.extend([1, 1, 1, 1, -4])
                    rows.extend([idx]*5)
                    cols.extend([lin_index(i, j-1), lin_index(i, j+1),
                                 lin_index(i-1, j), lin_index(i+1, j), idx])
        L = sp.coo_matrix((data, (rows, cols)), shape=(N, N)).tocsr()
        return L

    def solve(self) -> np.ndarray:
        """
        Solve the steady Navier–Stokes equations (penalty formulation) using FVM.
        
        For each velocity component (u and v), we solve:
            -ν Δ u + λ u = f_u,    -ν Δ v + λ v = f_v,
        where Δ is approximated via a five-point stencil.
        
        Dirichlet boundary conditions override the update in boundary cells.
        The two systems are solved independently, and the solution is returned as a concatenated vector.
        
        Returns:
            np.ndarray: The computed velocity field as a vector [u; v] of length 2*N.
        """
        # Grid parameters.
        n_x, n_y = self.grid_shape
        x_min, x_max = self.mesh.bounds[0]
        y_min, y_max = self.mesh.bounds[1]
        h = (x_max - x_min) / n_x  # assume square cells

        # Assemble discrete Laplacian operator L using FDM/FVM (five-point stencil).
        L = self._assemble_operator(self.grid_shape, h)

        # For each velocity component, the discrete equation is:
        #   -ν L u + λ u = f_u (times h^2 scaling from integration)
        # So, the system matrix for u is: A_u = -ν L + λ I.
        N = self.num_cells
        I = sp.identity(N)
        A_u = -self.nu * L + self.penalty * I
        A_v = -self.nu * L + self.penalty * I

        # Assemble right-hand side vectors.
        b_u = np.zeros(N)
        b_v = np.zeros(N)
        # For each control volume, evaluate f_u and f_v at the cell center.
        for i in range(N):
            center = self.cell_centers[i]
            b_u[i] = (h**2) * self.f_u(center)
            b_v[i] = (h**2) * self.f_v(center)

        # Apply Dirichlet boundary conditions on velocity components.
        # For each cell whose center lies on the boundary, override the corresponding row.
        tol = 1e-8
        bounds = self.mesh.bounds
        for i, center in enumerate(self.cell_centers):
            on_boundary = False
            for d, (lower, upper) in enumerate(bounds):
                if np.isclose(center[d], lower, atol=tol) or np.isclose(center[d], upper, atol=tol):
                    on_boundary = True
                    break
            if on_boundary:
                # Override for u.
                val_u = self.dirichlet_bc_u.func(center, 0)
                A_u.data[A_u.indptr[i]:A_u.indptr[i+1]] = 0
                A_u[i, i] = 1.0
                b_u[i] = val_u
                # Override for v.
                val_v = self.dirichlet_bc_v.func(center, 0)
                A_v.data[A_v.indptr[i]:A_v.indptr[i+1]] = 0
                A_v[i, i] = 1.0
                b_v[i] = val_v

        # (Optionally, add Neumann BC contributions if provided.)
        if self.neumann_bc_u is not None:
            bc_flux_u = self.neumann_bc_u.apply(self.mesh, 0)
            for idx, flux in bc_flux_u.items():
                b_u[idx] += flux
        if self.neumann_bc_v is not None:
            bc_flux_v = self.neumann_bc_v.apply(self.mesh, 0)
            for idx, flux in bc_flux_v.items():
                b_v[idx] += flux

        # Solve the linear systems.
        u_sol = spla.spsolve(A_u, b_u)
        v_sol = spla.spsolve(A_v, b_v)

        # Concatenate the solution (or return as a 2-column array).
        sol = np.hstack([u_sol.reshape(-1,1), v_sol.reshape(-1,1)])
        self.solution = sol
        return sol

# =============================================================================
# 2. Navier–Stokes Solver via Finite Difference Method (FDM)
# =============================================================================

class NavierStokesFDM:
    def __init__(self,
                 mesh: StructuredMesh,
                 nu: float,
                 penalty: float,
                 f_u: Callable[[np.ndarray], float],
                 f_v: Callable[[np.ndarray], float],
                 dirichlet_bc_u: DirichletCondition,
                 dirichlet_bc_v: DirichletCondition,
                 neumann_bc_u: Optional[NeumannCondition] = None,
                 neumann_bc_v: Optional[NeumannCondition] = None):
        """
        Initialize the FDM solver for the steady Navier–Stokes equations (penalty formulation).

        (Same parameters as in the FVM solver.)
        """
        self.mesh = mesh
        self.nu = nu
        self.penalty = penalty
        self.f_u = f_u
        self.f_v = f_v
        self.dirichlet_bc_u = dirichlet_bc_u
        self.dirichlet_bc_v = dirichlet_bc_v
        self.neumann_bc_u = neumann_bc_u
        self.neumann_bc_v = neumann_bc_v

        # For FDM, we assume the grid points are at the cell centers.
        self.cell_centers, self.cell_volumes, self.grid_shape = self._compute_cell_info()
        self.num_cells = self.cell_centers.shape[0]
        self.solution = None

    def _compute_cell_info(self):
        """Same as in the FVM solver."""
        if not hasattr(self.mesh, 'num_cells'):
            raise ValueError("Mesh must have 'num_cells' attribute.")
        n_x, n_y = self.mesh.num_cells
        x_min, x_max = self.mesh.bounds[0]
        y_min, y_max = self.mesh.bounds[1]
        dx = (x_max - x_min) / n_x
        dy = (y_max - y_min) / n_y
        centers = []
        volumes = []
        for i in range(n_x):
            for j in range(n_y):
                center = np.array([x_min + (i + 0.5)*dx, y_min + (j + 0.5)*dy])
                centers.append(center)
                volumes.append(dx*dy)
        return np.array(centers), np.array(volumes), (n_x, n_y)

    def _assemble_operator(self):
        """
        Assemble a 5-point stencil matrix for the Laplacian on the grid (FDM).
        For each interior point, the discrete Laplacian is approximated as
            (u_E + u_W + u_N + u_S - 4 u_P)/h^2.
        Returns the matrix L (in sparse CSR format).
        """
        n_x, n_y = self.grid_shape
        N = n_x * n_y
        data = []
        rows = []
        cols = []
        def lin_index(i, j):
            return i * n_y + j
        for i in range(n_x):
            for j in range(n_y):
                idx = lin_index(i, j)
                if i == 0 or i == n_x - 1 or j == 0 or j == n_y - 1:
                    # Boundary: placeholder.
                    data.append(1.0)
                    rows.append(idx)
                    cols.append(idx)
                else:
                    data.extend([1,1,1,1,-4])
                    rows.extend([idx]*5)
                    cols.extend([lin_index(i, j-1), lin_index(i, j+1),
                                 lin_index(i-1, j), lin_index(i+1, j), idx])
        L = sp.coo_matrix((data, (rows, cols)), shape=(N, N)).tocsr()
        return L

    def solve(self) -> np.ndarray:
        """
        Solve the steady Navier–Stokes equations (penalty formulation) using FDM.
        For each velocity component, the discrete equation is:
            -ν L u + λ u = f,
        where L is the Laplacian operator.
        Dirichlet BCs are enforced by replacing boundary rows with identity rows.
        The two systems for u and v are solved independently.
        
        Returns:
            np.ndarray: A two-column array containing the computed velocity field.
        """
        n_x, n_y = self.grid_shape
        N = self.num_cells
        L = self._assemble_operator()
        I = sp.identity(N)
        A_u = -self.nu * L + self.penalty * I
        A_v = -self.nu * L + self.penalty * I

        # Grid spacing (assumed uniform)
        x_min, x_max = self.mesh.bounds[0]
        h = (x_max - x_min) / n_x

        b_u = np.zeros(N)
        b_v = np.zeros(N)
        for i in range(N):
            center = self.cell_centers[i]
            b_u[i] = - (h**2) * self.f_u(center)
            b_v[i] = - (h**2) * self.f_v(center)

        tol = 1e-8
        bounds = self.mesh.bounds
        for i, center in enumerate(self.cell_centers):
            on_boundary = False
            for d, (lower, upper) in enumerate(bounds):
                if np.isclose(center[d], lower, atol=tol) or np.isclose(center[d], upper, atol=tol):
                    on_boundary = True
                    break
            if on_boundary:
                val_u = self.dirichlet_bc_u.func(center, 0)
                A_u.data[A_u.indptr[i]:A_u.indptr[i+1]] = 0
                A_u[i, i] = 1.0
                b_u[i] = val_u

                val_v = self.dirichlet_bc_v.func(center, 0)
                A_v.data[A_v.indptr[i]:A_v.indptr[i+1]] = 0
                A_v[i, i] = 1.0
                b_v[i] = val_v

        if self.neumann_bc_u is not None:
            bc_flux_u = self.neumann_bc_u.apply(self.mesh, 0)
            for i, flux in bc_flux_u.items():
                b_u[i] += flux
        if self.neumann_bc_v is not None:
            bc_flux_v = self.neumann_bc_v.apply(self.mesh, 0)
            for i, flux in bc_flux_v.items():
                b_v[i] += flux

        u_sol = spla.spsolve(A_u, b_u)
        v_sol = spla.spsolve(A_v, b_v)

        sol = np.hstack([u_sol.reshape(-1,1), v_sol.reshape(-1,1)])
        self.solution = sol
        return sol

# =============================================================================
# 3. Navier–Stokes Solver via Finite Element Method (FEM)
# =============================================================================

class NavierStokesFEM:
    def __init__(self,
                 mesh: StructuredMesh,
                 nu: float,
                 penalty: float,
                 f_u: Callable[[np.ndarray], float],
                 f_v: Callable[[np.ndarray], float],
                 dirichlet_bc_u: DirichletCondition,
                 dirichlet_bc_v: DirichletCondition,
                 neumann_bc_u: Optional[NeumannCondition] = None,
                 neumann_bc_v: Optional[NeumannCondition] = None):
        """
        Initialize the FEM solver for the steady Navier–Stokes equations (penalty formulation).

        Parameters:
            mesh (StructuredMesh): A structured mesh with attributes 'nodes' and 'cells'.
            nu (float): Kinematic viscosity.
            penalty (float): Penalty parameter to enforce near-incompressibility.
            f_u, f_v (Callable): Source term functions for the momentum equations.
            dirichlet_bc_u,v (DirichletCondition): BC objects for u and v.
            neumann_bc_u,v (Optional[NeumannCondition]): Optional Neumann BC objects.
        """
        self.mesh = mesh
        self.nu = nu
        self.penalty = penalty
        self.f_u = f_u
        self.f_v = f_v
        self.dirichlet_bc_u = dirichlet_bc_u
        self.dirichlet_bc_v = dirichlet_bc_v
        self.neumann_bc_u = neumann_bc_u
        self.neumann_bc_v = neumann_bc_v

        self.num_nodes = self.mesh.nodes.shape[0]
        self.solution = None
        # FEM assembly is performed on a per-element basis.
    
    def assemble_matrices(self):
        """
        Assemble the global stiffness matrices and load vectors for the velocity components
        using linear triangular elements.

        For each element (triangle), we compute:
          - The area via the determinant.
          - The gradients of the linear shape functions.
          - The local stiffness matrix: 
                Ke = area * (grad_N^T @ grad_N)
          - The local mass matrix:
                Me = (area/12)*(ones + I)
          - The local load vector is computed using centroid quadrature.
          - The global system for u is: (-nu*Ke + penalty*Me) u = f_u, similarly for v.
        """
        rows_A_u, cols_A_u, data_A_u = [], [], []
        rows_A_v, cols_A_v, data_A_v = [], [], []
        b_u = np.zeros(self.num_nodes)
        b_v = np.zeros(self.num_nodes)

        for cell in self.mesh.cells:
            coords = self.mesh.nodes[cell]  # (3,2)
            x = coords[:,0]
            y = coords[:,1]
            # Compute area of triangle.
            area = 0.5 * np.abs(np.linalg.det(np.array([[1, x[0], y[0]],
                                                         [1, x[1], y[1]],
                                                         [1, x[2], y[2]]])))
            centroid = np.mean(coords, axis=0)
            f_val_u = self.f_u(centroid)
            f_val_v = self.f_v(centroid)
            # Compute shape function gradients.
            b_vec = np.array([y[1]-y[2], y[2]-y[0], y[0]-y[1]])
            c_vec = np.array([x[2]-x[1], x[0]-x[2], x[1]-x[0]])
            grad_N = np.vstack((b_vec, c_vec)) / (2 * area)  # (2,3)
            Ke = area * (grad_N.T @ grad_N)  # Diffusion contribution.
            Me = (area / 12.0) * (np.ones((3,3)) + np.eye(3))
            Ke_penalized = -self.nu * Ke + self.penalty * Me

            # Local load vectors.
            fe_u = (area/3.0) * f_val_u * np.ones(3)
            fe_v = (area/3.0) * f_val_v * np.ones(3)

            # Assemble into global system.
            for i_local, i_global in enumerate(cell):
                b_u[i_global] += fe_u[i_local]
                b_v[i_global] += fe_v[i_local]
                for j_local, j_global in enumerate(cell):
                    rows_A_u.append(i_global)
                    cols_A_u.append(j_global)
                    data_A_u.append(Ke_penalized[i_local, j_local])
                    rows_A_v.append(i_global)
                    cols_A_v.append(j_global)
                    data_A_v.append(Ke_penalized[i_local, j_local])
        A_u = sp.coo_matrix((data_A_u, (rows_A_u, cols_A_u)), shape=(self.num_nodes, self.num_nodes)).tocsr()
        A_v = sp.coo_matrix((data_A_v, (rows_A_v, cols_A_v)), shape=(self.num_nodes, self.num_nodes)).tocsr()
        self.A_u = A_u
        self.A_v = A_v
        self.b_u = b_u
        self.b_v = b_v

    def apply_boundary_conditions(self):
        """
        Apply Dirichlet (and optionally Neumann) BCs for both velocity components.
        For each node on the boundary, override the corresponding row in the system and set the load.
        """
        tol = 1e-8
        bounds = self.mesh.bounds
        for i, coord in enumerate(self.mesh.nodes):
            on_boundary = False
            for d, (lower, upper) in enumerate(bounds):
                if np.isclose(coord[d], lower, atol=tol) or np.isclose(coord[d], upper, atol=tol):
                    on_boundary = True
                    break
            if on_boundary:
                # For u:
                val_u = self.dirichlet_bc_u.func(coord, 0)
                self.A_u.data[self.A_u.indptr[i]:self.A_u.indptr[i+1]] = 0
                self.A_u[i, i] = 1.0
                self.b_u[i] = val_u
                # For v:
                val_v = self.dirichlet_bc_v.func(coord, 0)
                self.A_v.data[self.A_v.indptr[i]:self.A_v.indptr[i+1]] = 0
                self.A_v[i, i] = 1.0
                self.b_v[i] = val_v

        # Optionally add Neumann contributions.
        if self.neumann_bc_u is not None:
            bc_flux_u = self.neumann_bc_u.apply(self.mesh, 0)
            for i, flux in bc_flux_u.items():
                self.b_u[i] += flux
        if self.neumann_bc_v is not None:
            bc_flux_v = self.neumann_bc_v.apply(self.mesh, 0)
            for i, flux in bc_flux_v.items():
                self.b_v[i] += flux

    def solve(self) -> np.ndarray:
        """
        Solve the steady Navier–Stokes equations (penalty formulation) using FEM.

        This method assembles the global system for the velocity components, applies boundary conditions,
        and then solves the linear systems to compute the velocity field.

        Returns:
            np.ndarray: A two-column array with the computed velocity field (u and v).
        """
        self.assemble_matrices()
        self.apply_boundary_conditions()
        u_sol = spla.spsolve(self.A_u, self.b_u)
        v_sol = spla.spsolve(self.A_v, self.b_v)
        sol = np.hstack([u_sol.reshape(-1,1), v_sol.reshape(-1,1)])
        self.solution = sol
        return sol