# pyfoam\pyfoam\core\solvers\fem_heat_solver.py

import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spla
from typing import Callable, Optional
from meshing.mesh import StructuredMesh
from conditions.condition import DirichletCondition, NeumannCondition

class HeatFEMSolver:
    def __init__(self,
                 mesh: StructuredMesh,
                 f: Callable[[np.ndarray, float], float],
                 dirichlet_bc: DirichletCondition,
                 neumann_bc: Optional[NeumannCondition] = None,
                 u0: Callable[[np.ndarray], float] = None):
        """
        Initialize the Heat FEM solver.

        Parameters:
            mesh (StructuredMesh): The structured mesh with attributes:
                                   - nodes: NumPy array of shape (N, dim).
                                   - cells: List of cells (each cell is a list of node indices for a triangle).
            f (Callable[[np.ndarray, float], float]): Source term function f(x,t).
            dirichlet_bc (DirichletCondition): Dirichlet boundary condition object g_D(x,t).
            neumann_bc (Optional[NeumannCondition]): Optional Neumann boundary condition object h_N(x,t).
            u0 (Callable[[np.ndarray], float]): Initial condition function u₀(x,y).
                                              Must be provided for time-dependent problems.
        """
        self.mesh = mesh
        self.f = f
        self.dirichlet_bc = dirichlet_bc
        self.neumann_bc = neumann_bc
        if u0 is None:
            raise ValueError("An initial condition u₀ must be provided for the heat equation.")
        self.u0 = u0

        self.num_nodes = self.mesh.nodes.shape[0]
        # Global stiffness matrix (A) and mass matrix (M)
        self.A = None
        self.M = None
        # Final solution (for time-dependent problems, solution history is stored)
        self.solution = None

    def assemble_matrices(self) -> None:
        """
        Assemble the global stiffness matrix A and the consistent mass matrix M.

        For each triangular element, the following local matrices are computed:
          - The area is computed via the determinant method.
          - The local stiffness matrix is:
                Ke = area * (grad_N^T @ grad_N)
            (Here we assume a constant diffusion coefficient of 1.)
          - The local consistent mass matrix is:
                Me = (area/12) * (ones(3,3) + I)
        The local matrices are assembled into global matrices A and M (stored in CSR format).
        """
        rows_A, cols_A, data_A = [], [], []
        rows_M, cols_M, data_M = [], [], []

        for cell in self.mesh.cells:
            coords = self.mesh.nodes[cell]  # shape: (3,2)
            x = coords[:, 0]
            y = coords[:, 1]
            # Compute area using the determinant formula.
            area = 0.5 * np.abs(np.linalg.det(np.array([[1, x[0], y[0]],
                                                         [1, x[1], y[1]],
                                                         [1, x[2], y[2]]])))
            centroid = np.mean(coords, axis=0)
            # Compute gradients of the linear shape functions.
            b_vec = np.array([y[1] - y[2], y[2] - y[0], y[0] - y[1]])
            c_vec = np.array([x[2] - x[1], x[0] - x[2], x[1] - x[0]])
            grad_N = np.vstack((b_vec, c_vec)) / (2 * area)  # shape: (2,3)
            # Local stiffness matrix (diffusion term).
            Ke = area * (grad_N.T @ grad_N)
            # Local consistent mass matrix.
            Me = (area / 12) * (np.ones((3, 3)) + np.eye(3))

            for i_local, i_global in enumerate(cell):
                for j_local, j_global in enumerate(cell):
                    rows_A.append(i_global)
                    cols_A.append(j_global)
                    data_A.append(Ke[i_local, j_local])
                    rows_M.append(i_global)
                    cols_M.append(j_global)
                    data_M.append(Me[i_local, j_local])
        A = sp.coo_matrix((data_A, (rows_A, cols_A)), shape=(self.num_nodes, self.num_nodes))
        M = sp.coo_matrix((data_M, (rows_M, cols_M)), shape=(self.num_nodes, self.num_nodes))
        self.A = A.tocsr()
        self.M = M.tocsr()

    def assemble_load_vector(self, t: float) -> np.ndarray:
        """
        Assemble the global load vector b at time t.

        For each element, the local load vector is computed using centroid quadrature:
            fe = (area/3) * f(centroid, t) * [1,1,1]^T
        and the contributions are summed into the global load vector.

        Parameters:
            t (float): Time at which to evaluate the source term.

        Returns:
            np.ndarray: Global load vector b (length = number of nodes).
        """
        b = np.zeros(self.num_nodes)
        for cell in self.mesh.cells:
            coords = self.mesh.nodes[cell]
            x = coords[:, 0]
            y = coords[:, 1]
            area = 0.5 * np.abs(np.linalg.det(np.array([[1, x[0], y[0]],
                                                         [1, x[1], y[1]],
                                                         [1, x[2], y[2]]])))
            centroid = np.mean(coords, axis=0)
            fe = (area / 3) * self.f(centroid, t) * np.ones(3)
            for i_local, i_global in enumerate(cell):
                b[i_global] += fe[i_local]
        return b

    def apply_boundary_conditions(self, b: np.ndarray, t: float) -> np.ndarray:
        """
        Apply boundary conditions to the system at time t.

        The DirichletCondition object's apply() method returns a dictionary mapping node indices
        to the prescribed value g_D(x, t). For each such node, the corresponding row in the global
        stiffness matrix is zeroed out and the diagonal entry is set to 1, and the load vector b is set
        to g_D(x, t). Similarly, the NeumannCondition (if provided) returns a dictionary mapping node indices
        to flux contributions h_N(x, t) which are added to b.

        Parameters:
            b (np.ndarray): Global load vector.
            t (float): Current time.

        Returns:
            np.ndarray: Modified load vector b with boundary conditions enforced.
        """
        # Apply Dirichlet conditions.
        bc_dirichlet = self.dirichlet_bc.apply(self.mesh, t)
        for node_idx, value in bc_dirichlet.items():
            self.A.data[self.A.indptr[node_idx]:self.A.indptr[node_idx+1]] = 0
            self.A[node_idx, node_idx] = 1.0
            b[node_idx] = value
        # Apply Neumann conditions if provided.
        if self.neumann_bc is not None:
            bc_neumann = self.neumann_bc.apply(self.mesh, t)
            for node_idx, flux in bc_neumann.items():
                b[node_idx] += flux
        return b

    def solve_time(self, dt: float, T: float) -> np.ndarray:
        """
        Solve the heat equation using implicit Euler time-stepping.

        The time-discrete system is:
            (M/dt + A) u^(n+1) = (M/dt)*u^(n) + b^(n+1),
        where b^(n+1) is the load vector (including source term and Neumann contributions)
        at time t^(n+1). Dirichlet boundary conditions are enforced at each time step.

        Parameters:
            dt (float): Time step size.
            T (float): Final time.

        Returns:
            np.ndarray: A 2D array of shape (num_steps+1, num_nodes) containing the solution
                        at each time step (each row is the flattened solution vector at that time).
        """
        num_steps = int(T / dt)
        if self.A is None or self.M is None:
            self.assemble_matrices()

        # Preassemble the constant left-hand side matrix.
        LHS = self.M / dt + self.A
        LHS = LHS.tocsr()

        # Initialize solution using the initial condition u₀.
        u = np.array([self.u0(x) for x in self.mesh.nodes])
        u = self.apply_boundary_conditions(u, t=0)

        solution_time = np.zeros((num_steps + 1, self.num_nodes))
        solution_time[0, :] = u.copy()

        t = 0.0
        for n in range(num_steps):
            t_next = t + dt
            b_time = self.assemble_load_vector(t_next)
            b_time = self.apply_boundary_conditions(b_time, t_next)
            RHS = self.M.dot(u) / dt + b_time
            u_new = spla.spsolve(LHS, RHS)
            u = u_new.copy()
            solution_time[n + 1, :] = u
            t = t_next

        self.solution = u
        return solution_time

    def solve(self) -> np.ndarray:
        """
        Generic solve method.

        For a time-dependent problem, please call solve_time(dt, T).
        This method is intended for stationary problems.

        Returns:
            np.ndarray: Computed nodal solution vector.
        """
        raise NotImplementedError("For the heat equation (time-dependent problem), please use solve_time(dt, T).")