# pyfoam\pyfoam\core\solvers\fem_poisson_solver.py

import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spla
from typing import Callable, Optional
from meshing.mesh import StructuredMesh
from conditions.condition import DirichletCondition, NeumannCondition

class PoissonFEMSolver:
    def __init__(self,
                 mesh: StructuredMesh,
                 f: Callable[[np.ndarray], float],
                 dirichlet_bc: DirichletCondition,
                 neumann_bc: Optional[NeumannCondition] = None):
        """
        Initialize the Poisson FEM solver.

        Parameters:
            mesh (StructuredMesh): A structured mesh with attributes:
                                   - nodes: a NumPy array of shape (N, dim) with node coordinates.
                                   - cells: a list of cells (each cell is a list of node indices for a triangle).
            f (Callable[[np.ndarray], float]): The source term function f(x,y).
            dirichlet_bc (DirichletCondition): Boundary condition object prescribing u = g(x,y) on ∂Ω.
            neumann_bc (Optional[NeumannCondition]): Optional boundary condition object prescribing flux h(x,y) on ∂Ω.
        """
        self.mesh = mesh
        self.f = f
        self.dirichlet_bc = dirichlet_bc
        self.neumann_bc = neumann_bc

        # Total number of nodes.
        self.num_nodes = self.mesh.nodes.shape[0]
        # Global stiffness matrix and load vector (to be assembled).
        self.A = None
        self.b = None
        # Final computed solution.
        self.solution = None

    def assemble_system(self) -> None:
        """
        Assemble the global stiffness matrix A and load vector b for the Poisson equation.

        For each triangular cell, the local stiffness matrix is computed using linear shape functions.
        The local stiffness matrix for a triangle is given by

            Ke = area * (grad_N^T @ grad_N)

        and the local load vector is computed via centroid quadrature as

            fe = (area/3) * f(centroid) * [1, 1, 1]^T.

        The local contributions are assembled into global matrices using row‐major ordering.
        """
        rows_A, cols_A, data_A = [], [], []
        b = np.zeros(self.num_nodes)

        for cell in self.mesh.cells:
            # Obtain the coordinates of the triangle's vertices.
            coords = self.mesh.nodes[cell]  # shape: (3,2)
            x = coords[:, 0]
            y = coords[:, 1]
            # Compute the area using the determinant formula.
            area = 0.5 * np.abs(np.linalg.det(np.array([[1, x[0], y[0]],
                                                         [1, x[1], y[1]],
                                                         [1, x[2], y[2]]])))
            # Compute the centroid.
            centroid = np.mean(coords, axis=0)
            # Evaluate the source term at the centroid.
            f_val = self.f(centroid)
            # Compute the gradients of the linear shape functions.
            # For a triangle with vertices (x_i, y_i), the shape function gradients are computed by:
            #   grad(N_i) = [b_i, c_i] / (2*area)
            b_vec = np.array([y[1] - y[2], y[2] - y[0], y[0] - y[1]])
            c_vec = np.array([x[2] - x[1], x[0] - x[2], x[1] - x[0]])
            grad_N = np.vstack((b_vec, c_vec)) / (2 * area)  # shape: (2,3)
            # Local stiffness matrix (diffusion term only, as c(x)=0 for Poisson).
            Ke = area * (grad_N.T @ grad_N)
            # Local load vector.
            fe = (area / 3) * f_val * np.ones(3)

            # Assemble local matrices into global system.
            for i_local, i_global in enumerate(cell):
                b[i_global] += fe[i_local]
                for j_local, j_global in enumerate(cell):
                    rows_A.append(i_global)
                    cols_A.append(j_global)
                    data_A.append(Ke[i_local, j_local])
        # Build the global stiffness matrix in CSR format.
        A = sp.coo_matrix((data_A, (rows_A, cols_A)), shape=(self.num_nodes, self.num_nodes))
        self.A = A.tocsr()
        self.b = b

    def apply_boundary_conditions(self) -> None:
        """
        Apply boundary conditions to the system.

        The DirichletCondition object's apply() method returns a dictionary mapping node indices
        to the prescribed value g(x,y). For each such node, the corresponding row in the global stiffness
        matrix is zeroed out and the diagonal entry is set to 1, and the load vector is set to g(x,y).

        If a NeumannCondition is provided, its contributions are added to the load vector.
        """
        # Apply Dirichlet BC.
        bc_dirichlet = self.dirichlet_bc.apply(self.mesh, t=0)
        for node_idx, value in bc_dirichlet.items():
            self.A.data[self.A.indptr[node_idx]:self.A.indptr[node_idx+1]] = 0
            self.A[node_idx, node_idx] = 1.0
            self.b[node_idx] = value
        # Apply Neumann BC if available.
        if self.neumann_bc is not None:
            bc_neumann = self.neumann_bc.apply(self.mesh, t=0)
            for node_idx, flux in bc_neumann.items():
                self.b[node_idx] += flux

    def solve(self) -> np.ndarray:
        """
        Solve the stationary Poisson equation.

        This method assembles the system, applies boundary conditions, and then
        solves the linear system A u = b using a sparse direct solver.

        Returns:
            np.ndarray: The computed nodal solution vector.
        """
        self.assemble_system()
        self.apply_boundary_conditions()
        u = spla.spsolve(self.A, self.b)
        self.solution = u
        return u