# pyfoam\pyfoam\core\solvers\fvm_helmholtz_solver.py

import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spla
from typing import Callable, Optional
from meshing.mesh import StructuredMesh
from conditions.condition import DirichletCondition, NeumannCondition

class HelmholtzFVMSolver:
    def __init__(self,
                 mesh: StructuredMesh,
                 f: Callable[[np.ndarray], float],
                 k: float,
                 dirichlet_bc: DirichletCondition,
                 neumann_bc: Optional[NeumannCondition] = None):
        """
        Initialize the Helmholtz FVM solver.

        Parameters:
            mesh (StructuredMesh): A structured mesh with attributes:
                                   - nodes: NumPy array of node coordinates.
                                   - cells: List of cells (each cell is a list of vertex indices).
                                   Also, the mesh must have 'bounds' and 'num_cells'.
            f (Callable[[np.ndarray], float]): Source term function f(x,y).
            k (float): Wave number.
            dirichlet_bc (DirichletCondition): Dirichlet boundary condition object prescribing u = g(x,y) on ∂Ω.
            neumann_bc (Optional[NeumannCondition]): Optional Neumann boundary condition object prescribing flux h(x,y) on ∂Ω.
        """
        self.mesh = mesh
        self.f = f
        self.k = k
        self.dirichlet_bc = dirichlet_bc
        self.neumann_bc = neumann_bc

        # For FVM, the degrees of freedom are the cell averages.
        self.cell_centers, self.cell_volumes, self.grid_shape = self._compute_cell_info()
        self.num_cells = self.cell_centers.shape[0]
        self.solution = None

    def _compute_cell_info(self):
        """
        Compute cell centers and volumes for a structured rectangular mesh.

        For a 2D mesh with bounds [(x_min, x_max), (y_min, y_max)] and num_cells = [n_x, n_y]:
            dx = (x_max - x_min)/n_x, dy = (y_max - y_min)/n_y.
        The center of cell (i,j) is:
            (x_min + (i+0.5)*dx, y_min + (j+0.5)*dy),
        and the cell area is dx * dy.

        Returns:
            cell_centers (np.ndarray): Array of shape (n_cells, 2) with cell center coordinates.
            cell_volumes (np.ndarray): Array of cell areas.
            grid_shape (tuple): (n_x, n_y)
        """
        if not hasattr(self.mesh, 'num_cells'):
            raise ValueError("Mesh must have 'num_cells' attribute.")
        n_x, n_y = self.mesh.num_cells
        x_min, x_max = self.mesh.bounds[0]
        y_min, y_max = self.mesh.bounds[1]
        dx = (x_max - x_min) / n_x
        dy = (y_max - y_min) / n_y

        centers = []
        volumes = []
        for i in range(n_x):
            for j in range(n_y):
                center = np.array([x_min + (i + 0.5) * dx, y_min + (j + 0.5) * dy])
                centers.append(center)
                volumes.append(dx * dy)
        return np.array(centers), np.array(volumes), (n_x, n_y)

    def _assemble_system(self) -> (sp.csr_matrix, np.ndarray):
        """
        Assemble the global linear system for the Helmholtz equation using FVM.

        For a uniform rectangular grid with cell size h (assume h = dx = dy), the discrete
        equation for an interior cell is:
        
            u_E + u_W + u_N + u_S - (4 - k^2 h^2) u_P = - h^2 f_P,
        
        where f_P is the source term evaluated at the cell center.
        
        Returns:
            A (sp.csr_matrix): Global coefficient matrix.
            b (np.ndarray): Global load vector.
        """
        n_x, n_y = self.grid_shape
        x_min, x_max = self.mesh.bounds[0]
        y_min, y_max = self.mesh.bounds[1]
        dx = (x_max - x_min) / n_x
        dy = (y_max - y_min) / n_y
        h = dx  # assume square cells

        N = self.num_cells  # total number of cells
        A_data = []
        A_rows = []
        A_cols = []
        b = np.zeros(N)

        # Helper function: convert 2D index (i,j) to linear index.
        def lin_index(i, j):
            return i * n_y + j

        # Loop over each cell.
        for i in range(n_x):
            for j in range(n_y):
                idx = lin_index(i, j)
                center = self.cell_centers[idx]
                f_val = self.f(center)
                # For interior cells, use five-point stencil.
                if 0 < i < n_x - 1 and 0 < j < n_y - 1:
                    # Off-diagonals: neighbor cells get coefficient +1.
                    A_data.extend([1, 1, 1, 1])
                    A_rows.extend([idx] * 4)
                    A_cols.extend([lin_index(i, j-1), lin_index(i, j+1),
                                    lin_index(i-1, j), lin_index(i+1, j)])
                    # Central cell: coefficient -(4 - k^2 h^2).
                    A_data.append(-(4 - self.k**2 * h**2))
                    A_rows.append(idx)
                    A_cols.append(idx)
                    b[idx] = - h**2 * f_val
                else:
                    # For boundary cells, set up a placeholder (to be replaced by BC).
                    A_data.append(1.0)
                    A_rows.append(idx)
                    A_cols.append(idx)
                    b[idx] = - h**2 * f_val
        A = sp.coo_matrix((A_data, (A_rows, A_cols)), shape=(N, N)).tocsr()
        return A, b

    def apply_boundary_conditions(self, A: sp.csr_matrix, b: np.ndarray) -> (sp.csr_matrix, np.ndarray):
        """
        Apply Dirichlet and (optionally) Neumann boundary conditions to the global system.

        For each cell whose center is on the boundary (within a tolerance), the DirichletCondition's
        function is used to override the cell value. This is done by replacing the corresponding row
        in A with an identity row and setting b to the prescribed value.
        Neumann conditions (if provided) are added to b.

        Parameters:
            A (sp.csr_matrix): Global coefficient matrix.
            b (np.ndarray): Global load vector.

        Returns:
            A, b: The modified matrix and load vector.
        """
        tol = 1e-8
        bounds = self.mesh.bounds
        for idx, center in enumerate(self.cell_centers):
            on_boundary = False
            for d, (lower, upper) in enumerate(bounds):
                if np.isclose(center[d], lower, atol=tol) or np.isclose(center[d], upper, atol=tol):
                    on_boundary = True
                    break
            if on_boundary and (self.dirichlet_bc is not None):
                value = self.dirichlet_bc.func(center, 0)  # time not used in stationary Helmholtz
                A.data[A.indptr[idx]:A.indptr[idx+1]] = 0
                A[idx, idx] = 1.0
                b[idx] = value
        if self.neumann_bc is not None:
            bc_flux = self.neumann_bc.apply(self.mesh, 0)
            for idx, flux in bc_flux.items():
                b[idx] += flux
        return A, b

    def solve(self) -> np.ndarray:
        """
        Solve the stationary Helmholtz equation using FVM.

        The method assembles the linear system, applies boundary conditions, and then
        solves A u = b for the cell-average values.

        Returns:
            np.ndarray: The computed cell-average solution vector.
        """
        A, b = self._assemble_system()
        A, b = self.apply_boundary_conditions(A, b)
        u = spla.spsolve(A, b)
        self.solution = u
        return u